/* $Id: header 297205 2010-03-30 21:09:07Z johannes $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "ext/mysqlnd/mysqlnd.h"
#include "php_mysqlnd_server_locator.h"
#include "Zend/zend_ini.h"
#include "Zend/zend_ini_scanner.h"

ZEND_DECLARE_MODULE_GLOBALS(mysqlnd_server_locator)

static int plugin_id;
static func_mysqlnd_conn__connect orig_mysqlnd_conn_connect_method;

static HashTable server_list;
static time_t server_list_init = 0;

#ifdef ZTS
MUTEX_T LOCK_init_server_list;
#endif

#define INI_FILE_PATH "/tmp/server_list.ini"

static void mysqlnd_server_locator_ini_parser_cb(zval *arg1, zval *arg2, zval *arg3, int callback_type, void *list_v TSRMLS_DC) /* {{{ */
{
	HashTable *list = (HashTable*)list_v;
	char *hostname;

	if (!arg1 || !arg2) {
		return;
	}

	switch (callback_type)
	{
		case ZEND_INI_PARSER_ENTRY:
			hostname = pestrndup(Z_STRVAL_P(arg2), Z_STRLEN_P(arg2), 1);
			zend_hash_update(list, Z_STRVAL_P(arg1), Z_STRLEN_P(arg1) + 1, &hostname, sizeof(char *), NULL);
			break;
		case ZEND_INI_PARSER_SECTION:
			break;
		case ZEND_INI_PARSER_POP_ENTRY:
			php_error_docref(NULL TSRMLS_CC, E_NOTICE, "Array syntax not allowed in ini file");
			break;
		default:
			php_error_docref(NULL TSRMLS_CC, E_NOTICE, "Unexpected callback_type while parsing server list ini file");
			break;
	}
}
/* }}} */

static int mysqlnd_server_locator_init_server_list(TSRMLS_D) /* {{{ */
{
	zend_file_handle fh;

	memset(&fh, 0, sizeof(fh));
	fh.filename = MYSL_G(ini_file);
	fh.type = ZEND_HANDLE_FILENAME;

	if (zend_parse_ini_file(&fh, 0, ZEND_INI_SCANNER_NORMAL, mysqlnd_server_locator_ini_parser_cb, &server_list TSRMLS_CC) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failed to parse server list ini file");
		return FAILURE;
	}

	return SUCCESS;
}
/* }}} */


/* {{{ mysqlnd_server_locator::connect */
static enum_func_status MYSQLND_METHOD(mysqlnd_server_locator, connect)(MYSQLND * conn,
		const char *host, const char *user,
		const char *passwd, unsigned int passwd_len,
		const char *db, unsigned int db_len,
		unsigned int port,
		const char * socket_or_pipe,
		unsigned int mysql_flags
		TSRMLS_DC)
{
	char **new_host;
	char *actual_host = host;
	php_stream_statbuf ssb;


#ifdef ZTS
	tsrm_mutex_lock(LOCK_init_server_list);
#endif

	if (php_stream_stat_path(MYSL_G(ini_file), &ssb) == SUCCESS) {
		if (ssb.sb.st_mtime > server_list_init) {
			zend_hash_clean(&server_list);
			server_list_init = 0;
		}
	}

	if (!server_list_init) {
		mysqlnd_server_locator_init_server_list(TSRMLS_C);
		server_list_init = ssb.sb.st_mtime;
	}

#ifdef ZTS
	tsrm_mutex_unlock(LOCK_init_server_list);
#endif

	if (zend_hash_find(&server_list, user, strlen(user) + 1, (void**)&new_host) == SUCCESS) {
		actual_host = *new_host;
	}

	return orig_mysqlnd_conn_connect_method(conn, actual_host, user, passwd, passwd_len, db, db_len, port, socket_or_pipe, mysql_flags TSRMLS_CC);
}
/* }}} */

/* {{{ PHP_INI */
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("mysqlnd_server_locator.ini_file", INI_FILE_PATH, PHP_INI_SYSTEM, OnUpdateStringUnempty, ini_file, zend_mysqlnd_server_locator_globals, mysqlnd_server_locator_globals)
PHP_INI_END()
/* }}} */

/* {{{ php_mysqlnd_server_locator_init_globals
 */
static void php_mysqlnd_server_locator_init_globals(zend_mysqlnd_server_locator_globals *mysqlnd_server_locator_globals)
{
	mysqlnd_server_locator_globals->ini_file = NULL;
}
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(mysqlnd_server_locator)
{
	struct st_mysqlnd_conn_methods *conn_methods;
	ZEND_INIT_MODULE_GLOBALS(mysqlnd_server_locator, php_mysqlnd_server_locator_init_globals, NULL);
	REGISTER_INI_ENTRIES();

#ifdef ZTS
	LOCK_init_server_list = tsrm_mutex_alloc();
#endif

	plugin_id = mysqlnd_plugin_register();
	conn_methods = mysqlnd_conn_get_methods();

	orig_mysqlnd_conn_connect_method = conn_methods->connect;
	conn_methods->connect = MYSQLND_METHOD(mysqlnd_server_locator, connect);

	if (zend_hash_init(&server_list, 10, NULL, free, 1) == FAILURE) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failed to init server_list table");
		return FAILURE;
	}   

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(mysqlnd_server_locator)
{
	UNREGISTER_INI_ENTRIES();
	zend_hash_destroy(&server_list);
	tsrm_mutex_free(LOCK_init_server_list);

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(mysqlnd_server_locator)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "mysqlnd_server_locator support", "enabled");
	php_info_print_table_end();

	/* Remove comments if you have entries in php.ini
	DISPLAY_INI_ENTRIES();
	*/
}
/* }}} */


static const zend_module_dep mysqlnd_server_locator_deps[] = {
	ZEND_MOD_REQUIRED("mysqlnd")
	{NULL, NULL, NULL}
};

/* {{{ mysqlnd_server_locator_module_entry
 */
zend_module_entry mysqlnd_server_locator_module_entry = {
	STANDARD_MODULE_HEADER_EX,
	NULL,
	mysqlnd_server_locator_deps,
	"mysqlnd_server_locator",
	NULL,
	PHP_MINIT(mysqlnd_server_locator),
	PHP_MSHUTDOWN(mysqlnd_server_locator),
	NULL,
	NULL,
	PHP_MINFO(mysqlnd_server_locator),
	"0.1",
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_MYSQLND_SERVER_LOCATOR
ZEND_GET_MODULE(mysqlnd_server_locator)
#endif


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
